use crate::{
    apub::{
        actions::{
            apub::require_not_blocked, CreateServer, CreateServerApub, DeleteServer, UpdateServer,
        },
        ExtendedApplication,
    },
    recover, Action, Context, Error, RecoverableError, Required,
};
use activitystreams::prelude::*;
use url::Url;

pub mod follow;

pub(crate) fn application(
    application: &ExtendedApplication,
    key_owner: Option<&Url>,
    ctx: &Context,
) -> Result<Result<Box<dyn Action>, RecoverableError>, Error> {
    let id = application.id_unchecked().req()?;
    require_not_blocked(key_owner, id, ctx)?;

    // Double create
    if ctx.apub.object(id)?.is_some() {
        return Err(Error::Invalid);
    }

    let handle = application.preferred_username().req()?.to_owned();
    let domain = id.domain().req()?.to_owned();

    if handle != domain {
        return Err(Error::Invalid);
    }

    let title = application
        .name()
        .and_then(|n| n.one())
        .and_then(|s| s.as_xsd_string())
        .map(|s| s.to_owned());
    let description = application
        .summary()
        .and_then(|s| s.as_single_xsd_string())
        .map(|s| s.to_owned());
    let public_key = application.ext_one.public_key.public_key_pem.clone();
    let public_key_id = application.ext_one.public_key.id.clone();
    let published = application.published().req()?.into();

    Ok(Ok(Box::new(CreateServer {
        apub: Some(CreateServerApub {
            application_apub_id: id.to_owned(),
            public_key_id,
            public_key,
        }),
        domain,
        title,
        description,
        published,
    })))
}

pub(crate) fn create_application(
    create: &activitystreams::activity::Create,
    application_obj: &ExtendedApplication,
    key_owner: Option<&Url>,
    ctx: &Context,
) -> Result<Result<Box<dyn Action>, RecoverableError>, Error> {
    let actor_id = create.actor()?.as_single_id().req()?;
    require_not_blocked(key_owner, actor_id, ctx)?;

    let application_id = application_obj.id_unchecked().req()?;

    if actor_id != application_id {
        return Err(Error::Invalid);
    }

    application(application_obj, None, ctx)
}

pub(crate) fn update_application(
    update: &activitystreams::activity::Update,
    application: &ExtendedApplication,
    key_owner: Option<&Url>,
    ctx: &Context,
) -> Result<Result<Box<dyn Action>, RecoverableError>, Error> {
    let update_actor = update.actor()?.as_single_id().req()?;
    require_not_blocked(key_owner, update_actor, ctx)?;

    let id = application.id_unchecked().req()?;

    if update_actor != id {
        return Err(Error::Invalid);
    }

    let handle = application.preferred_username().req()?.to_owned();
    let domain = id.domain().req()?.to_owned();

    if handle != domain {
        return Err(Error::Invalid);
    }

    let server_id = recover!(id, ctx.apub.id_for_apub(id)?);
    let server_id = server_id.server().req()?;

    let title = application
        .name()
        .and_then(|n| n.one())
        .and_then(|s| s.as_xsd_string())
        .map(|s| s.to_owned());
    let description = application
        .summary()
        .and_then(|s| s.as_single_xsd_string())
        .map(|s| s.to_owned());
    let public_key_id = application.ext_one.public_key.id.clone();
    let public_key = application.ext_one.public_key.public_key_pem.clone();

    Ok(Ok(Box::new(UpdateServer {
        server_id,
        title,
        description,
        public_key: Some((public_key_id, public_key)),
    })))
}

pub(crate) fn delete_application(
    delete: &activitystreams::activity::Delete,
    application: &ExtendedApplication,
    key_owner: Option<&Url>,
    ctx: &Context,
) -> Result<Result<Box<dyn Action>, RecoverableError>, Error> {
    let delete_actor = delete.actor()?.as_single_id().req()?;
    require_not_blocked(key_owner, delete_actor, ctx)?;

    let id = application.id_unchecked().req()?;

    if delete_actor != id {
        return Err(Error::Invalid);
    }

    let server_id = recover!(id, ctx.apub.id_for_apub(id)?);
    let server_id = server_id.server().req()?;

    Ok(Ok(Box::new(DeleteServer { server_id })))
}
