use super::{ServerCreated, ServerUpdated};
use crate::{
    apub::{ExtendedApplication, PublicKey, PublicKeyInner},
    Context, Error, Outbound, Required,
};
use activitystreams::{
    activity::{Create, Update},
    actor::{ApActor, Application, Endpoints},
    base::AnyBase,
    context,
    prelude::*,
    security,
};
use std::collections::HashSet;
use url::Url;
use uuid::Uuid;

impl Outbound for ServerCreated {
    fn id(&self) -> Option<Uuid> {
        Some(self.server_id)
    }

    fn inboxes(&self, _: &Context) -> Result<Vec<Url>, Error> {
        Ok(vec![])
    }

    fn to_apub(&self, ctx: &Context) -> Result<AnyBase, Error> {
        let server = ctx.store.servers.by_id(self.server_id)?.req()?;

        let application_id = ctx.apub.info.gen_id().req()?;
        let public_key_id = ctx.apub.info.public_key(&application_id).req()?;
        let following = ctx.apub.info.following(&application_id).req()?;
        let followers = ctx.apub.info.followers(&application_id).req()?;
        let inbox = ctx.apub.info.inbox(&application_id).req()?;
        let outbox = ctx.apub.info.outbox(&application_id).req()?;
        let shared_inbox = ctx.apub.info.shared_inbox();

        ctx.apub.server(&application_id, server.id())?;
        ctx.apub.store_server_endpoints(
            server.id(),
            &application_id,
            crate::apub::Endpoints {
                inbox: inbox.clone(),
                outbox: outbox.clone(),
                following: following.clone(),
                followers: followers.clone(),
                shared_inbox: shared_inbox.clone(),
                public_key: public_key_id.clone(),
            },
        )?;
        let public_key_pem = ctx.apub.gen_server_keys(server.id(), &application_id)?;

        let mut application = ExtendedApplication::new(
            ApActor::new(inbox, Application::new()),
            PublicKey {
                public_key: PublicKeyInner {
                    id: public_key_id,
                    owner: application_id.clone(),
                    public_key_pem,
                },
            },
        );

        application
            .set_id(application_id.clone())
            .set_following(following)
            .set_followers(followers)
            .set_outbox(outbox)
            .set_endpoints(Endpoints {
                shared_inbox: Some(shared_inbox),
                ..Endpoints::default()
            })
            .set_preferred_username(server.domain())
            .set_published(server.created().into());

        if let Some(title) = server.title() {
            application.set_name(title);
        }
        if let Some(description) = server.description() {
            application.set_summary(description);
        }

        let any_base = application.into_any_base()?;
        ctx.apub.store_object(&any_base)?;

        let mut create = Create::new(application_id, any_base);
        create
            .set_id(ctx.apub.info.gen_id().req()?)
            .add_context(context())
            .add_context(security());

        let any_base = create.into_any_base()?;
        ctx.apub.store_object(&any_base)?;

        Ok(any_base)
    }
}

impl Outbound for ServerUpdated {
    fn id(&self) -> Option<Uuid> {
        Some(self.server_id)
    }

    fn inboxes(&self, ctx: &Context) -> Result<Vec<Url>, Error> {
        let follower_inboxes = ctx
            .store
            .federated_servers()
            .filter_map(|server_id| {
                Some(ctx.apub.endpoints_for_server(server_id).ok()??.shared_inbox)
            })
            .collect::<HashSet<_>>();

        Ok(follower_inboxes.into_iter().collect())
    }

    fn to_apub(&self, ctx: &Context) -> Result<AnyBase, Error> {
        let server = ctx.store.servers.by_id(self.server_id)?.req()?;
        let endpoints = ctx.apub.endpoints_for_server(self.server_id)?.req()?;
        let public_key_id = ctx.apub.key_for_server(server.id())?.req()?;
        let public_key_pem = ctx.apub.public_key_for_id(&public_key_id)?.req()?;
        let application_id = ctx.apub.apub_for_server(self.server_id)?.req()?;

        let mut application = ExtendedApplication::new(
            ApActor::new(endpoints.inbox, Application::new()),
            PublicKey {
                public_key: PublicKeyInner {
                    id: public_key_id,
                    owner: application_id.clone(),
                    public_key_pem,
                },
            },
        );

        application
            .set_id(application_id.clone())
            .set_following(endpoints.following)
            .set_followers(endpoints.followers)
            .set_outbox(endpoints.outbox)
            .set_endpoints(Endpoints {
                shared_inbox: Some(endpoints.shared_inbox),
                ..Endpoints::default()
            })
            .set_preferred_username(server.domain())
            .set_published(server.created().into());

        if let Some(title) = server.title() {
            application.set_name(title);
        }
        if let Some(description) = server.description() {
            application.set_summary(description);
        }

        let any_base = application.into_any_base()?;
        ctx.apub.store_object(&any_base)?;

        let mut update = Update::new(application_id, any_base);
        update
            .set_id(ctx.apub.info.gen_id().req()?)
            .add_context(context())
            .add_context(security());

        let any_base = update.into_any_base()?;
        ctx.apub.store_object(&any_base)?;

        Ok(any_base)
    }
}
